/*
  A simple server program.  It sets up a TCP port for the client
  program to connect to.

  This version is unthreaded and very slow if there is any I/O delay
  simulated.
*/

/*
cc -o server server.c -L. -R. -g -lpthread -lthread -pthread_extensions -lposix4 -lnsl -lsocket
*/

/* #define _POSIX_C_SOURCE 199506L */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/uio.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <time.h>
#include "thread_extensions.h"



struct request_struct
{int			socket_fd;
 char 			*buf;
 pthread_t		tid;
 struct request_struct	*next;
};
typedef struct request_struct request_t;

pthread_attr_t		attr;
int 			TCP_PORT = 6500;
int 			SLEEP = 10;
int 			SPIN = 0;
int			N_THREADS = 0;




void count_requests(int i)		/* Note the encapsulation */
{static int count=0, old_count=0, max_count = 0;

 count += i;
 if (abs(count - old_count) > 999)
   {printf("Processed %d requests.\n", count);
    old_count = count;}
 if (count > max_count)
   max_count = count;
}  


void *process_request(void *arg)
{char obuf[100];
 request_t *request = (request_t *) arg;
 pthread_t tid = pthread_self();

 DEBUG(printf("[%s] received: \"%s\"\n", thread_name(tid), request->buf));
 delay(SLEEP, SPIN);
 sprintf(obuf, "Server[%s] %s", thread_name(tid), request->buf);
 write(request->socket_fd, obuf, 50);
 free(request->buf);
 free(request);
}


request_t *get_request(int socket_fd)
{request_t *request;
 char 	ibuf[100];

 read(socket_fd, ibuf, 50);
 if (strcmp(ibuf, "End") == 0) return(NULL);

 request = (request_t *) malloc(sizeof(request_t));
 request->socket_fd = socket_fd;
 request->buf = (char *) malloc(strlen(ibuf)+1);
 strcpy(request->buf, ibuf);
 count_requests(1);
 return(request);
}


void producer(int socket_fd)
{request_t *request;

 while (1)
   {request = get_request(socket_fd);
    if (request == NULL) return;
      process_request(request);
  }
}


void *accept_requests(void *arg)
{int 	socket_fd = (int) arg;
 char 	ibuf[100], obuf[100];
 int 	i, j;
 pthread_t tid = pthread_self();

 printf("[%s] Accepting on socket: %d\n", thread_name(tid), socket_fd);

 producer(socket_fd);
 write(socket_fd, "End", 4);
 close(socket_fd);
 printf("[%s] Done Processing.    ", thread_name(tid));
}

void *stopper(void *arg)
{
  sleep(30);
  exit(NULL);
}

main(int argc, char **argv)
{int i,	sockfd, newsockfd, clilen;
 struct sockaddr_in cli_addr, serv_addr;

 if (argc >= 2) TCP_PORT = atoi(argv[1]);
 if (argc >= 3) SLEEP = atoi(argv[2]);
 if (argc >= 4) SPIN = atoi(argv[3]);
 printf("TCP_PORT = %d SLEEP (ms) = %d SPIN (us) = %d N_THREADS = %d\n",
	TCP_PORT, SLEEP, SPIN, N_THREADS);

 if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   fprintf(stderr,"server: can't open stream socket\n"), exit(0);
 memset((char *) &serv_addr, 0, sizeof(serv_addr));
 serv_addr.sin_family = AF_INET;
 serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
 serv_addr.sin_port = htons(TCP_PORT);
 if(bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
   fprintf(stderr,"server: can't bind local address\n"), exit(0);
 listen(sockfd, 5);
    
 PTHREAD_ATTR_INIT(&attr);
 pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
 pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);



 while (1)
   {clilen = sizeof(cli_addr);
    newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
    if(newsockfd < 0)
      fprintf(stderr,"server: accept error\n"), exit(0);
    accept_requests((void *)newsockfd);
  }
}
